"""This moduel contains argument validation for classes.

Most validations are for public arguments of the 'Dedupe' class.

However, some validations exist for other private classes
"""

from collections import Counter
from typing import Literal

from pyspark.sql import SparkSession

from liken._constants import INVALID_COLUMNS_EMPTY
from liken._constants import INVALID_COLUMNS_NOT_NONE
from liken._constants import INVALID_COLUMNS_REPEATED
from liken._constants import INVALID_DEDUPER
from liken._constants import INVALID_DF
from liken._constants import INVALID_KEEP
from liken._constants import INVALID_PREPROCESSOR
from liken._constants import INVALID_SPARK
from liken._constants import SUPPORTED_DFS
from liken._dedupers import BaseDeduper
from liken._preprocessors import Preprocessor
from liken._types import Columns
from liken._types import UserDataFrame


def validate_df_arg(df: UserDataFrame) -> UserDataFrame:
    if not isinstance(df, SUPPORTED_DFS):
        raise ValueError(INVALID_DF.format(type(df).__name__))
    return df


def validate_spark_arg(spark_session: SparkSession | None = None, /) -> SparkSession:
    """Validates Spark arg in the 'Dedupe' class"""
    if not spark_session:
        raise ValueError(INVALID_SPARK)
    return spark_session


def validate_keep_arg(keep: Literal["first", "last"]) -> Literal["first", "last"]:
    """Validates Keep arg in the 'Dedupe' class"""
    # TODO: do a type check and TypeError raise here too
    if keep not in ("first", "last"):
        raise ValueError(INVALID_KEEP.format(keep))
    return keep


def validate_deduper_arg(deduper: BaseDeduper) -> BaseDeduper:
    """Validates that the given 'deduper' is in fact a `BaseDeduper`.

    As used by the collections manager
    """
    if not isinstance(deduper, BaseDeduper):
        raise TypeError(INVALID_DEDUPER.format(type(deduper).__name__))
    return deduper


def validate_columns_arg(
    columns: Columns | None,
    is_sequential_applied: bool,
) -> Columns | None:
    """validates inputs to public api 'columns' arg.

    Allowed combinations are:

    - Sequential API: .canonicalize with columns defined
    - Dict API: .canonicalize with NO columns defined
    - Pipeline API: .canonicalize with NO columns defined

    Any other combination/repetion raises a value error
    """
    if is_sequential_applied:
        if not columns:
            raise ValueError(INVALID_COLUMNS_EMPTY)

        if isinstance(columns, tuple):
            for label, count in Counter(
                columns,
            ).items():
                if count > 1:
                    raise ValueError(INVALID_COLUMNS_REPEATED.format(label))

    if not is_sequential_applied and columns:
        raise ValueError(INVALID_COLUMNS_NOT_NONE)
    return columns


def validate_preprocessor_arg(preprocessor: Preprocessor) -> Preprocessor:
    """Validates that the given arg is in fact a `Preprocessor`"""
    if not isinstance(preprocessor, Preprocessor):
        raise TypeError(INVALID_PREPROCESSOR.format(type(preprocessor).__name__))
    return preprocessor



import warnings
from typing import final


# EXCEPTIONS:


@final
class InvalidDeduperError(TypeError):
    def __init__(self, msg):
        super().__init__(msg)


# TODO: these warnings come up was "UserWarning". Change.
def warn(msg: str) -> None:
    return warnings.warn(msg, category=UserWarning)



from liken import custom
from liken import preprocessors
from liken import synthetic
from liken._dedupers import cosine
from liken._dedupers import exact
from liken._dedupers import fuzzy
from liken._dedupers import isin
from liken._dedupers import isna
from liken._dedupers import jaccard
from liken._dedupers import lsh
from liken._dedupers import str_contains
from liken._dedupers import str_endswith
from liken._dedupers import str_len
from liken._dedupers import str_startswith
from liken._dedupers import tfidf
from liken._pipelines import Col
from liken._pipelines import Pipeline
from liken._pipelines import col
from liken._pipelines import pipeline
from liken.liken import Dedupe
from liken.liken import dedupe


__all__ = [
    "dedupe",
    "Dedupe",
    "exact",
    "fuzzy",
    "lsh",
    "tfidf",
    "cosine",
    "jaccard",
    "isin",
    "isna",
    "str_contains",
    "str_endswith",
    "str_len",
    "str_startswith",
    "pipeline",
    "Pipeline",
    "col",
    "Col",
    "custom",
    "preprocessors",
    "synthetic",
]



"""Define custom dedupers"""

from functools import wraps
from typing import Callable

from liken._custom import Custom
from liken._custom import PairGenerator
from liken._registry import registry


def register(f: PairGenerator) -> Callable:
    """Register a custom function as a deduper.

    Custom functions can be registered for use as dedupers recognised by the
    ``Dedupe`` class. Use ``register`` as a decorator around the custom callable.

    The custom callable must accept a generic array-like object representing the
    contents of one or more DataFrame columns. The concrete column backing this
    array is resolved only when the deduper is applied.

    The expected function signature is:

        function(array, **kwargs)

    Args:
        f: A custom callable that returns integer pairs of indices identifying
            similar pairs in an array. Accepted callables are functions or
            generators, where generators are preferred.

    Returns:
        Callable

    Raises:
        TypeError: If any positional arguments are used when calling the
            registered deduper.

    Example:
        Registering a custom deduper

            import liken as lk

            @lk.custom.register
            def custom_deduper(array, **kwargs):
                # your code here
                yield ...

            df = (
                lk.dedupe(df)
                .apply(custom_deduper(**kwargs))
                .drop_duplicates("address")
                .collect()
            )

        E.g. the following Custom exact string-length deduplication deduper:

            @lk.custom.register
            def eq_str_len(array):
                n = len(array)
                for i in range(n):
                    for j in range(i + 1, n):
                        if len(array[i]) == len(array[j]):
                            yield i, j

        Applying the deduper:

            df = (
                lk.dedupe(df)
                .apply(eq_str_len()) # array arg implicitely passed
                .drop_duplicates("address")
                .collect()
            )

        Before:

            +------+-----------+
            | id   | address   |
            +------+-----------+
            |  1   | london    |
            |  2   | paris     |
            |  3   | tokyo     |
            +------+-----------+

        "tokyo" and "paris" have the same length, so reduced:

            +------+-----------+
            | id   | address   |
            +------+-----------+
            |  1   | london    |
            |  2   | paris     |
            +------+-----------+

        Keyword-only enforcement:

            Deduper(df).apply(my_func(is_upper_caps=True))  # OK
            Deduper(df).apply(my_func(True))                # Raises TypeError
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        if args:
            raise TypeError(f"{f.__name__} must be called with keyword arguments only")
        return Custom(f, **kwargs)

    # Add to registry
    registry.register(f"{f.__name__}", func=wrapper)

    return wrapper



"""Deduplication collectionexecutors.

`SparkExecutor` simply calls a partition processor where each partition will
then be processed with the `LocalExecutor`
"""

# mypy: disable-error-code="no-redef"

from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterator
from functools import partial
from typing import TYPE_CHECKING
from typing import Protocol
from typing import Type
from typing import TypeAlias
from typing import TypeVar
from typing import cast
from typing import final

from networkx.utils.union_find import UnionFind
from pyspark.sql import Row
from pyspark.sql import SparkSession

from liken._collections import SEQUENTIAL_API_DEFAULT_KEY
from liken._collections import DeduplicationDict
from liken._collections import Pipeline
from liken._constants import CANONICAL_ID
from liken._dataframe import Frame
from liken._dataframe import LocalDF
from liken._dataframe import SparkDF
from liken._dedupers import BaseDeduper
from liken._dedupers import PredicateDedupers
from liken._preprocessors import Preprocessor
from liken._types import Columns
from liken._types import Keep


if TYPE_CHECKING:
    from liken.liken import Dedupe


# TYPES:


SingleComponents: TypeAlias = dict[int, list[int]]
MultiComponents: TypeAlias = dict[tuple[int, ...], list[int]]
F = TypeVar("F", bound=Frame)


# EXECUTORS:


class Executor(Protocol[F]):
    def execute(
        self,
        df: F,
        /,
        *,
        columns: Columns | None,
        dedupers: DeduplicationDict | Pipeline,
        keep: Keep,
        drop_duplicates: bool,
        drop_canonical_id: bool,
        id: str | None,
    ) -> F: ...


@final
class LocalExecutor(Executor):
    def execute(
        self,
        df: LocalDF,
        /,
        *,
        columns: Columns | None,
        dedupers: DeduplicationDict | Pipeline,
        keep: Keep,
        drop_duplicates: bool,
        drop_canonical_id: bool,
        id: str | None = None,
    ) -> LocalDF:
        """Process a local dataframe according to the deduplication collection

        Processing is defined according to whether the collections of
        dedupers is a:
            - Pipeline: in which case "and" combination steps are allowed
            - DeduplicationDict: in which case handles Sequential and Dict API

        For Pipeline, predication is implemented if an and combination contains at
        least one Predicate Deduper. In that case the predicate dedupers are proxy
        WHERE filters that propagate a set of dataframe indice positions to the
        next deduper (most likely a threshold deduper, but optionally predicate
        also.)
        """

        del id  # Unused: here for interface symmetry with SparkExecutor

        call_deduper = partial(
            self._call_deduper,
            drop_duplicates=drop_duplicates,
            keep=keep,
        )

        if isinstance(dedupers, DeduplicationDict):
            if not columns:
                for col, iter_dedupers in dedupers.items():
                    for deduper in iter_dedupers:
                        uf, n = self._build_uf(deduper, df, col)
                        components: SingleComponents = self._get_components(uf, n)
                        df = call_deduper(deduper, components)
            else:
                # For sequential API calls e.g.`.canonicalize("address")`
                for deduper in dedupers[SEQUENTIAL_API_DEFAULT_KEY]:
                    uf, n = self._build_uf(deduper, df, columns)
                    components: SingleComponents = self._get_components(uf, n)
                    df = call_deduper(deduper, components)

        if isinstance(dedupers, Pipeline):
            for step in dedupers.steps:
                any_predicate: bool = dedupers._has_any_predicate(step)

                # predication only if at least one predicate deduper
                if any_predicate:
                    indices = set()

                    for col, deduper, preprocessor in step:
                        uf, n = self._build_uf(deduper, df, col, preprocessor, predicate=indices)

                        components = defaultdict(list)
                        idx: list = sorted(indices)
                        for i in range(n):
                            if not indices:
                                components[uf[i]].append(i)
                            else:
                                components[idx[uf[i]]].append(idx[i])

                        if isinstance(deduper, PredicateDedupers):
                            for c in components.values():
                                if len(c) > 1:
                                    indices = indices.union(set(c))

                else:
                    ufs = []

                    for col, deduper, preprocessor in step:
                        uf, n = self._build_uf(deduper, df, col, preprocessor)
                        ufs.append(uf)
                    components: MultiComponents = self._get_multi_components(ufs, n)

                df = call_deduper(deduper, components)

        if drop_canonical_id:
            return df.drop_col(CANONICAL_ID)
        return df

    @staticmethod
    def _build_uf(
        deduper: BaseDeduper,
        df: LocalDF,
        columns: Columns,
        preprocessors: list[Preprocessor] = [],
        predicate: set = set(),
    ) -> tuple[UnionFind[int], int]:
        return deduper.set_frame(df).build_union_find(columns, preprocessors, predicate=predicate)

    @staticmethod
    def _get_components(
        uf: UnionFind[int],
        n: int,
    ) -> SingleComponents:
        components = defaultdict(list)
        for i in range(n):
            components[uf[i]].append(i)
        return components

    @staticmethod
    def _get_multi_components(
        ufs: list[UnionFind[int]],
        n: int,
    ) -> MultiComponents:
        components = defaultdict(list)
        for i in range(n):
            signature = tuple(uf[i] for uf in ufs)
            components[signature].append(i)
        return components

    @staticmethod
    def _call_deduper(
        deduper: BaseDeduper,
        components: SingleComponents | MultiComponents,
        drop_duplicates: bool,
        keep: Keep,
    ) -> LocalDF:
        return deduper.canonicalizer(
            components=components,
            drop_duplicates=drop_duplicates,
            keep=keep,
        )


@final
class SparkExecutor(Executor):
    def __init__(self, spark_session: SparkSession, id: str | None = None):
        self._spark_session = spark_session
        self._id = id

    def execute(
        self,
        df: SparkDF,
        /,
        *,
        columns: Columns | None,
        dedupers: DeduplicationDict | Pipeline,
        keep: Keep,
        drop_duplicates: bool,
        drop_canonical_id: bool,
        id: str | None = None,
    ) -> SparkDF:
        """Spark specific deduplication helper

        Maps dataframe partitions to be processed via the RDD API yielding low-
        level list[Rows], which are then post-processed back to a dataframe.

        Args:
            columns: The attribute to deduplicate.
            dedupers: the collection of dedupers
        Retuns:
            Instance's _df attribute is updated
        """

        # import in worker node
        from liken.liken import Dedupe

        # IMPORTANT: Use local variables, no references to Self
        process_partition = self._process_partition

        rdd = df.mapPartitions(
            lambda partition: process_partition(
                factory=Dedupe,
                partition=partition,
                dedupers=dedupers,
                id=id,
                columns=columns,
                drop_duplicates=drop_duplicates,
                keep=keep,
            )
        )

        schema = df._schema

        df = SparkDF(self._spark_session.createDataFrame(rdd, schema=schema), is_init=False)

        if drop_canonical_id:
            return df.drop_col(CANONICAL_ID)
        return df

    @staticmethod
    def _process_partition(
        *,
        factory: Type[Dedupe],
        partition: Iterator[Row],
        dedupers: DeduplicationDict | Pipeline,
        id: str | None,
        columns: Columns | None,
        drop_duplicates: bool,
        keep: Keep = "first",
    ) -> Iterator[Row]:
        """process a spark dataframe partition i.e. a list[Row]

        This function is functionality mapped to a worker node. For clean
        separation from the driver, dedupers are re-instantiated and the main
        liken API is executed *per* worker node.

        Args:
            paritition_iter: a partition
            dedupers: the collection of dedupers
            id: the unique identified of the dataset a.k.a "business key"
            columns: the attribute on which to deduplicate

        Returns:
            A list[Row], deduplicated
        """
        # handle empty partitions
        rows: list[Row] = list(partition)
        if not rows:
            return iter([])

        # Core API reused per partition, per worker node
        df = (
            factory._from_rows(rows)
            .apply(dedupers)
            .canonicalize(
                columns,
                keep=keep,
                drop_duplicates=drop_duplicates,
                id=id,
            )
            .collect()
        )

        return iter(cast(list[Row], df))



import re
from typing import Literal

import nltk
import pyarrow as pa
import pyarrow.compute as pc
from cleanco import basename
from nameparser import HumanName
from nltk.corpus import stopwords as nltk_stopwords
from typing_extensions import override


# BASE:


class Preprocessor:
    """Base class for all preprocessors."""

    def from_array(self, array: pa.Array) -> None:
        """array setter"""
        self._array = array

    def process(self) -> pa.Array:
        raise NotImplementedError("Child class must implement process()")


# STRING PROCESSORS:


class Strip(Preprocessor):
    """Remove leading/trailing whitespace"""

    @override
    def process(self) -> pa.Array:
        return pc.utf8_trim_whitespace(self._array)


class Lower(Preprocessor):
    """Convert strings to lowercase"""

    @override
    def process(self) -> pa.Array:
        return pc.utf8_lower(self._array)


class Alnum(Preprocessor):
    """Remove non-alphanumeric characters, including spaces"""

    @override
    def process(self) -> pa.Array:
        return pc.replace_substring_regex(self._array, "[^0-9A-Za-z]+", "")


class RemovePunctuation(Preprocessor):
    """Remove punctuation but preserve spaces"""

    @override
    def process(self) -> pa.Array:
        return pc.replace_substring_regex(self._array, r"[^\w\s]+", "")


class NormalizeUnicode(Preprocessor):
    """Normalize Unicode strings"""

    def __init__(
        self,
        form: Literal["NFC", "NFKC", "NFD", "NFKD"] = "NFKD",
    ):
        self._form = form

    @override
    def process(self) -> pa.Array:
        return pc.utf8_normalize(self._array, form=self._form)


class AsciiFold(Preprocessor):
    """Converts alphabetic, numeric, and symbolic characters that are not in
    the Basic Latin Unicode block (first 127 ASCII characters) to their ASCII
    equivalent, if one exists. For example, the filter changes à to a.
    """

    def __init__(self, form: str = "NFKD"):
        self._form = form

    @override
    def process(self) -> pa.Array:
        self._array = pc.utf8_normalize(self._array, form="NFKD")

        return pc.replace_substring_regex(self._array, "[̀-ͯ]", "")


# STOPWORDS PROCESSOR:


class RemoveStopwords(Preprocessor):
    """Remove stopwords"""

    def __init__(self, words: list[str] | None = None, language: str = "english"):
        self._words = words
        self._language = language

    @override
    def process(self) -> pa.Array:
        if not self._words:
            try:
                nltk.data.find("corpora/stopwords")
            except LookupError:
                nltk.download("stopwords")
            self._words = nltk_stopwords.words(self._language)

        pattern = r"\b(" + "|".join(re.escape(w) for w in self._words) + r")\b"

        return pc.replace_substring_regex(self._array, pattern, "")


# NAME PROCESSORS:


class NormalizeName(Preprocessor):
    """Normalize personal names"""

    @override
    def process(self) -> pa.Array:
        def _clean_name(name: str) -> str:

            hn = HumanName(name)
            return f"{hn.first} {hn.middle} {hn.last}".strip()

        # Important: cast to python list first
        pylist = self._array.to_pylist()

        return pa.array([_clean_name(x) if x is not None else None for x in pylist])


class NormalizeCompany(Preprocessor):
    """Normalize company names"""

    @override
    def process(self) -> pa.Array:
        # Important: cast to python list first
        pylist = self._array.to_pylist()

        return pa.array([basename(x) if x is not None else None for x in pylist])



from liken._preprocessors import Alnum
from liken._preprocessors import AsciiFold
from liken._preprocessors import Lower
from liken._preprocessors import NormalizeCompany
from liken._preprocessors import NormalizeName
from liken._preprocessors import NormalizeUnicode
from liken._preprocessors import RemovePunctuation
from liken._preprocessors import RemoveStopwords
from liken._preprocessors import Strip


def strip() -> Strip:
    """Remove leading/trailing whitespace."""
    return Strip()


def lower() -> Lower:
    """Convert strings to lowercase."""
    return Lower()


def alnum() -> Alnum:
    """Remove non-alphanumeric characters, including spaces."""
    return Alnum()


def remove_punctuation() -> RemovePunctuation:
    """Remove punctuation but preserve spaces."""
    return RemovePunctuation()


def normalize_unicode(form: str = "NFKD") -> NormalizeUnicode:
    """Normalize Unicode strings.

    Args:
        form: Unicode normalization form. Accepted values are "NFC", "NFKC",
            "NFD", "NFKD".
    """
    return NormalizeUnicode(form=form)


def ascii_fold() -> AsciiFold:
    """Converts alphabetic, numeric, and symbolic characters that are not in
    the Basic Latin Unicode block (first 127 ASCII characters) to their ASCII
    equivalent, if one exists. For example, the filter changes à to a.
    """
    return AsciiFold()


def remove_stopwords(
    words: list[str] | None = None,
    language: str = "english",
) -> RemoveStopwords:
    """Remove stopwords.

    Args:
        words: A list of words to ignore. If defined, `language` argument is
            ignored.
        language: The language to use for the stop words dictionary"""
    return RemoveStopwords(words=words, language=language)


def normalize_names() -> NormalizeName:
    """Normalize personal names.

    Preserves only first name, middle name and last name. Titles and nicknames
    are stripped. Commas are cleaned.
    """
    return NormalizeName()


def normalize_company() -> NormalizeCompany:
    """Normalize company names.

    Strips common company name nomenclature e.g. "Ltd.", or "LLC".
    """
    return NormalizeCompany()



"""TODO"""

from __future__ import annotations

from typing import NamedTuple
from typing import Self
from typing import TypeAlias
from typing import final

from liken._dedupers import BaseDeduper
from liken._dedupers import PredicateDedupers
from liken._preprocessors import Preprocessor
from liken._registry import registry
from liken._types import Columns
from liken._validators import validate_preprocessor_arg


# TYPES:


class PipelineUnit(NamedTuple):
    columns: Columns
    deduper: BaseDeduper
    preprocessors: list[Preprocessor]


PipelineStep: TypeAlias = list[PipelineUnit]
PipelineCollection: TypeAlias = list[PipelineStep]

InputPreprocessor: TypeAlias = Preprocessor | list[Preprocessor]

# PUBLIC ON API:


def pipeline(preprocessors: InputPreprocessor = []) -> Pipeline:
    """Convenience function for `Pipeline` collection."""
    return Pipeline(preprocessors)


def col(columns: Columns, /, *, preprocessors: InputPreprocessor = []) -> Col:
    """Convenience function for calling `Col` with a deduper."""
    return Col(columns, preprocessors=preprocessors)


# ON:


@final
class Col:
    """Unit collection for a single deduper in the Pipeline API.

    Operates deduplication on a column. Passed as a step in a deduplication
    pipeline. Allows for usage of dedupers as method calls instead of functions.

    Args:
        columns: the label(s) of a column or columns.

    Returns:
        None

    Example:
        single `on` in a pipeline step:

            import liken as lk

            lk.pipeline().step(on("address").exact())

        Dedupers are combined as "and" statements when passed as a members
        of a list in a step:

            lk.pipeline().step(
                [
                    on("email").fuzzy(threshold=0.95),
                    on("email").str_endswith("UK"),
                ]
            )

        Dedupers can be combined as "and" statement in a step for
        **different** columns:

            lk.pipeline().step(
                [
                    on("email").fuzzy(threshold=0.95),
                    ~on("address").isna(),
                ]
            )

        Pipeline preprocessors will be applied to all dedupers:

            pipeline = (
                lk.pipeline(preprocessors=[lk.preprocessors.lower()])
                .step(lk.col("email").fuzzy()) # Preprocessor applies here
                .step(lk.col("address").tfidf()) # And here
            )

        But, a global pipeline preprocessor will not override an explicit
        deduper's preprocessor. Similarily, a step's preprocessor will not
        override the deduper's:

            pipeline = (
                lk.pipeline(preprocessors=[lk.preprocessors.ascii_fold()])
                .step(
                    [
                        lk.col("email").fuzzy(),  # preprocessed by step's preprocessor, `alnum`.
                        ~lk.col(
                            "address",
                            preprocessors=[lk.preprocessors.lower()],
                        ).isna(), # uses it's own preprocessor, `lower`.
                    ],
                    preprocessors=[lk.preprocessors.alnum()], # defines the step's preprocessor
                )
                .step(lk.col("address").tfidf()) # defaults to the pipeline's preprocessor, `ascii_fold`.
            )
    """

    # for IDE autocompletion only!
    # must be manually maintained
    # add a new dummy method here upon adding a new deduper.

    def exact(self, *args, **kwargs) -> Col:
        """Method wrapper of `lk.exact` function.

        Usage is identical to the function but chained to an instance of `Col`.
        See [`lk.exact`](../reference/liken.md#liken.exact) reference for
        complete documentation.

        Example:
            Define as part of a pipeline:

                pipeline = lk.pipeline().step(on("col").exact())

        """
        return self.__getattr__("exact")(*args, **kwargs)

    def fuzzy(self, *args, **kwargs) -> Col:
        """Method wrapper of `lk.fuzzy` function.

        Usage is identical to the function but chained to an instance of `Col`.
        See [`lk.fuzzy`](../reference/liken.md#liken.fuzzy) reference for
        complete documentation.

        Example:
            Define as part of a pipeline:

                pipeline = lk.pipeline().step(on("col").fuzzy())

        """
        return self.__getattr__("fuzzy")(*args, **kwargs)

    def tfidf(self, *args, **kwargs) -> Col:
        """Method wrapper of `lk.tfidf` function.

        Usage is identical to the function but chained to an instance of `Col`.
        See [`lk.tfidf`](../reference/liken.md#liken.tfidf) reference for
        complete documentation.

        Example:
            Define as part of a pipeline:

                pipeline = lk.pipeline().step(on("col").tfidf())

        """
        return self.__getattr__("tfidf")(*args, **kwargs)

    def lsh(self, *args, **kwargs) -> Col:
        """Method wrapper of `lk.lsh` function.

        Usage is identical to the function but chained to an instance of `Col`.
        See [`lk.lsh`](../reference/liken.md#liken.lsh) reference for
        complete documentation.

        Example:
            Define as part of a pipeline:

                pipeline = lk.pipeline().step(on("col").lsh())

        """
        return self.__getattr__("lsh")(*args, **kwargs)

    def jaccard(self, *args, **kwargs) -> Col:
        """Method wrapper of `lk.jaccard` function.

        Usage is identical to the function but chained to an instance of `Col`.
        See [`lk.jaccard`](../reference/liken.md#liken.jaccard) reference for
        complete documentation.

        Example:
            Define as part of a pipeline:

                pipeline = lk.pipeline().step(on("col").jaccard())

        """
        return self.__getattr__("jaccard")(*args, **kwargs)

    def cosine(self, *args, **kwargs) -> Col:
        """Method wrapper of `lk.cosine` function.

        Usage is identical to the function but chained to an instance of `Col`.
        See [`lk.cosine`](../reference/liken.md#liken.cosine) reference for
        complete documentation.

        Example:
            Define as part of a pipeline:

                pipeline = lk.pipeline().step(on("col").cosine())

        """
        return self.__getattr__("cosine")(*args, **kwargs)

    def isin(self, *args, **kwargs) -> Col:
        """Method wrapper of `lk.isin` function.

        Usage is identical to the function but chained to an instance of `Col`.
        See [`lk.isin`](../reference/liken.md#liken.isin) reference for
        complete documentation.

        Example:
            Define as part of a pipeline:

                pipeline = lk.pipeline().step(on("col").isin())

        """
        return self.__getattr__("isin")(*args, **kwargs)

    def isna(self, *args, **kwargs) -> Col:
        """Method wrapper of `lk.isna` function.

        Usage is identical to the function but chained to an instance of `Col`.
        See [`lk.isna`](../reference/liken.md#liken.isna) reference for
        complete documentation.

        Example:
            Define as part of a pipeline:

                pipeline = lk.pipeline().step(on("col").isna())

        """
        return self.__getattr__("isna")(*args, **kwargs)

    def str_startswith(self, *args, **kwargs) -> Col:
        """Method wrapper of `lk.str_startswith` function.

        Usage is identical to the function but chained to an instance of `Col`.
        See [`lk.str_startswith`](../reference/liken.md#liken.str_startswith) reference for
        complete documentation.

        Example:
            Define as part of a pipeline:

                pipeline = lk.pipeline().step(on("col").str_startswith())

        """
        return self.__getattr__("str_startswith")(*args, **kwargs)

    def str_endswith(self, *args, **kwargs) -> Col:
        """Method wrapper of `lk.str_endswith` function.

        Usage is identical to the function but chained to an instance of `Col`.
        See [`lk.str_endswith`](../reference/liken.md#liken.str_endswith) reference for
        complete documentation.

        Example:
            Define as part of a pipeline:

                pipeline = lk.pipeline().step(on("col").str_endswith())

        """
        return self.__getattr__("str_endswith")(*args, **kwargs)

    def str_contains(self, *args, **kwargs) -> Col:
        """Method wrapper of `lk.str_contains` function.

        Usage is identical to the function but chained to an instance of `Col`.
        See [`lk.str_contains`](../reference/liken.md#liken.str_contains) reference for
        complete documentation.

        Example:
            Define as part of a pipeline:

                pipeline = lk.pipeline().step(on("col").str_contains())

        """
        return self.__getattr__("str_contains")(*args, **kwargs)

    def str_len(self, *args, **kwargs) -> Col:
        """Method wrapper of `lk.str_len` function.

        Usage is identical to the function but chained to an instance of `Col`.
        See [`lk.str_len`](../reference/liken.md#liken.str_len) reference for
        complete documentation.

        Example:
            Define as part of a pipeline:

                pipeline = lk.pipeline().step(on("col").str_len())

        """
        return self.__getattr__("str_len")(*args, **kwargs)

    def __init__(self, columns: Columns, preprocessors: InputPreprocessor = []):
        self._columns: Columns = columns
        self._unit: PipelineUnit
        self._preprocessors: list[Preprocessor] = resolve_preprocessors(preprocessors)

    def __getattr__(self, attr):
        """Make deduper functions available as method calls to Col.

        Functions are retrieved from registry. Includes any prior custom
        dedupers that have been registered.
        """

        # No intercept: Python internals
        if attr.startswith("__"):
            raise AttributeError(attr)

        func = registry.get(f"{attr}")

        def wrapper(*args, **kwargs):
            deduper = func(*args, **kwargs)
            self._unit = PipelineUnit(self._columns, deduper, self._preprocessors)
            return self

        return wrapper

    def __invert__(self) -> Col:
        """Propagate inverstion to the deduper. Allows for following syntax:

        ~on("email").isna()

        Where the inversion get's propagated to act on isna().
        """

        columns, deduper, preprocessors = self._unit

        new_on = Col(columns)
        new_on._unit = PipelineUnit(columns, ~deduper, preprocessors)
        return new_on

    @property
    def unit(self) -> PipelineUnit:
        return self._unit

    def __str__(self) -> str:
        """string representation

        Parses a single Col or combinations of Col operated with `&`
        """
        on: str = "lk.col"

        rep = ""
        columns, deduper, _ = self._unit
        deduper: str = str(deduper)
        if deduper.startswith("~"):
            deduper = deduper[1:]
            on = "~" + on
        rep += f"{on}('{columns}').{deduper}"
        return rep


# PIPELINE:


class Pipeline:
    """Builds deduplication pipeline.

    Defines deduplication steps as chainable calls. Accepts preprocessors.
    Preprocessors are propagate to each step and on each column, unless
    explicit overriden in those stages.

    Args:
        preprocessors: a preprocessor, or list of preprocessors to apply to the
            pipeline

    Returns:
        None

    Raises:
        TypeError: if passed preprocessor not a member of `liken.preprocessors`
    """

    def __init__(self, preprocessors: InputPreprocessor = []):
        self._preprocessors: list[Preprocessor] = resolve_preprocessors(preprocessors)
        self._ons: list[list[Col]] = []
        self._steps: PipelineCollection = []

    def step(
        self,
        ons: Col | list[Col],
        /,
        *,
        preprocessors: InputPreprocessor = [],
    ) -> Self:
        """Define a deduplication pipeline step

        Accepts a single deduper or a list of dedupers. A list of dedupers that
        has more than one member are composed as "and rules" meaning that the
        deduplication of both (or more) dedupers are considered for
        canonicalisation.

        Args:
            ons: `Col` deduper, or list of the same.
            preprocessors: a preprocessor, or list of preprocessors to appy the
                whole step (i.e. to all dedupers if more than one deduper).

        Returns:
            Self

        Example:
            A single deduper is added:

                pipeline = lk.pipeline().step(lk.col("email").exact())

            Multiple steps can be chained:

                pipeline = (
                    lk.pipeline()
                    .step(lk.col("email").fuzzy())
                    .step(lk.col("address").tfidf())
                )

            More than on deduper can be added to form composable rules within
            a step:

                pipeline = (
                    lk.pipeline()
                    .step(
                        [
                            lk.col("email").fuzzy(),
                            ~lk.col("address").isna(),
                        ]
                    )
                    .step(lk.col("address").tfidf())
                )

            Pipeline preprocessors will be applied to all steps:

                pipeline = (
                    lk.pipeline(preprocessors=[lk.preprocessors.lower()])
                    .step(lk.col("email").fuzzy()) # Preprocessor applies here
                    .step(lk.col("address").tfidf()) # And here
                )

            But, a global pipeline preprocessor will not override an explicit
            step's preprocessor

                pipeline = (
                    lk.pipeline(preprocessors=[lk.preprocessors.lower()])
                    .step(
                        lk.col("email").fuzzy(), preprocessors=[lk.preprocessors.alnum()]
                    )  # only `alnum` preprocessed
                    .step(lk.col("address").tfidf())  # this one still preprocessed with `lower`
                )
        """
        preprocessors: list[Preprocessor] = resolve_preprocessors(preprocessors)
        if not preprocessors:
            preprocessors = self._preprocessors

        if isinstance(ons, Col):
            ons: list[Col] = [ons]
        self._ons.append(ons)

        step: PipelineStep = [on.unit for on in ons]

        # propagate preprocessor
        step = [s._replace(preprocessors=preprocessors) if not s.preprocessors else s for s in step]

        # predicates sorted to first for "rule predication"
        step: PipelineStep = sorted(step, key=lambda x: not isinstance(x[1], PredicateDedupers))

        self._steps.append(step)

        return self

    def __str__(self) -> str:
        pros = ""
        if preprocessors := self._preprocessors:
            pros = "preprocessors=" + f"{[str(p) for p in preprocessors]}"

        inner = ""
        for step in self._ons:
            inner += "\n\t\t" + ".step(["
            for on in step:
                inner += "\n\t\t\t" + str(on) + ","
            inner += "\n\t\t" + "])"
        return f"(\n\tlk.rules.builder({pros}){inner}\n)"

    @staticmethod
    def _has_any_predicate(step: list[tuple[Columns, BaseDeduper]]) -> bool:
        """Retrieve whether or not the pipeline step has at least one predicate
        deduper.

        Private! not public API.
        """
        return any([isinstance(x[1], PredicateDedupers) for x in step])

    @property
    def steps(self):
        return self._steps


# HELPERS:


def resolve_preprocessors(
    preprocessors: InputPreprocessor,
) -> list[Preprocessor]:
    if isinstance(preprocessors, list):
        return [validate_preprocessor_arg(p) for p in preprocessors]
    # i.e. a single preprocessor
    return [validate_preprocessor_arg(preprocessors)]



"""Defines collections of dedupers"""

from __future__ import annotations

from collections import UserDict
from copy import deepcopy
from typing import final

from liken._constants import INVALID_DICT_KEY_MSG
from liken._constants import INVALID_DICT_MEMBER_MSG
from liken._constants import INVALID_DICT_VALUE_MSG
from liken._constants import INVALID_FALLBACK_MSG
from liken._constants import INVALID_SEQUENCE_AFTER_DICT_MSG
from liken._constants import SEQUENTIAL_API_DEFAULT_KEY
from liken._constants import WARN_DICT_REPLACES_SEQUENCE_MSG
from liken._constants import WARN_RULES_REPLACES_RULES_MSG
from liken._dedupers import BaseDeduper
from liken._exceptions import InvalidDeduperError
from liken._exceptions import warn
from liken._pipelines import Col
from liken._pipelines import Pipeline


# DICT CONFIG:


@final
class DeduplicationDict(UserDict):
    """Dict collection for dedupers in the Sequential and Dict APIs

    For Sequential API all values (dedupers) are added under a default key.

    For Dict API column label(s) (i.e. str or tuple) are the keys."""

    def __setitem__(self, key, value):
        if not isinstance(key, str | tuple):
            raise InvalidDeduperError(INVALID_DICT_KEY_MSG.format(type(key).__name__))
        if not isinstance(value, list | tuple | BaseDeduper):
            raise InvalidDeduperError(INVALID_DICT_VALUE_MSG.format(type(value).__name__))
        if not isinstance(value, BaseDeduper):
            for i, member in enumerate(value):
                if not isinstance(member, BaseDeduper):
                    raise InvalidDeduperError(INVALID_DICT_MEMBER_MSG.format(i, key, type(member).__name__))
        else:
            value = (value,)
        super().__setitem__(key, value)

    def __str__(self):
        rep = ""
        for k, values in self.items():
            krep = ""
            for v in values:
                krep += "\n\t\t" + str(v) + ","
            rep += f"\n\t'{k}': ({krep[:-1]},\n\t\t),"
        return "{" + rep + "\n}"


# COLLECTIONS MANAGER:


@final
class CollectionsManager:
    """
    Manage and validate collection(s) of dedupers.

    Supports addition of dedupers as part of the three APIs:
    - Sequential
    - Dict
    - Pipeline

    For Sequential dedupers, as instances of `BaseDeduper` are sequentially
    to an idential structure of the Dict API but under a single default
    dictionary key. Keys are columns names, and values are iterables of
    dedupers.

    Raises:
        InvalidDeduperError for any misconfigured addition of a deduper.
    """

    def __init__(self) -> None:
        self._dedupers: DeduplicationDict | Pipeline = DeduplicationDict({SEQUENTIAL_API_DEFAULT_KEY: []})
        self.has_applies: bool = False

    @property
    def is_sequential_applied(self) -> bool:
        """checks to see if any dedupers exist in the default key"""
        if isinstance(self._dedupers, Pipeline):
            return False
        return set(self._dedupers) == {SEQUENTIAL_API_DEFAULT_KEY}

    def apply(self, deduper: BaseDeduper | dict | DeduplicationDict | Pipeline) -> None:
        """Loads a deduper / collection of dedupers into the manager

        This function currently handles all possible instances of dedupers, and
        the implementation achieves this by writing to the deduplication
        dictionary or overwriting the dictionary with `Pipeline`.

        If the input deduper is `BaseDeduper` then "Sequential" API is in use. If
        dict (or DeduplicationDict — even though this is not public) then it is the
        "Dict" API. Else "Pipeline" API is in use.

        Note also that as Pipeline contains Col and combinations of Col operated
        with & results in self mutation, need deep copy to allow for
        serialization to Spark workers."""

        # track that at least one apply made
        # if not, used by `Dedupe` to include an exact deduper by default
        self.has_applies = True

        if isinstance(deduper, BaseDeduper):
            if not self.is_sequential_applied:
                raise InvalidDeduperError(INVALID_SEQUENCE_AFTER_DICT_MSG)
            self._dedupers[SEQUENTIAL_API_DEFAULT_KEY].append(deduper)
            return

        if isinstance(deduper, dict | DeduplicationDict):
            if self._dedupers[SEQUENTIAL_API_DEFAULT_KEY]:
                warn(WARN_DICT_REPLACES_SEQUENCE_MSG)
            self._dedupers = DeduplicationDict(deduper)
            return

        if isinstance(deduper, Col):
            deduper = Pipeline().step(deduper)

        if isinstance(deduper, Pipeline):
            if isinstance(self._dedupers, Pipeline):
                warn(WARN_RULES_REPLACES_RULES_MSG)
            # required for spark serialization
            self._dedupers = deepcopy(deduper)  # type: ignore
            return

        raise InvalidDeduperError(INVALID_FALLBACK_MSG.format(type(deduper).__name__))

    def get(self) -> DeduplicationDict | Pipeline:
        return self._dedupers

    def pretty_get(self) -> None | str:
        """string representation of dedupers.

        Output string must be formatted approximately such that it can be used
        with .apply(), i.e. a string representation of one of:
            - BaseDeduper
            - DeduplicationDict
            - Pipeline
        The seuqneital API with numerous additions of BaseStraegy means there
        is not good way to retried this such that is available to "apply". So,
        default to returning it as a list representation.
        """
        dedupers = self.get()

        if isinstance(dedupers, DeduplicationDict):
            if self.is_sequential_applied:
                deduper: list = dedupers[SEQUENTIAL_API_DEFAULT_KEY]
                if not deduper:
                    return None
                return str(*deduper)
            return str(dedupers)
        return str(dedupers)

    def reset(self):
        """Reset collection to empty defaultdict"""
        self._dedupers = DeduplicationDict({SEQUENTIAL_API_DEFAULT_KEY: []})



"""liken main public API"""

from __future__ import annotations

from typing import Self

import pandas as pd
import polars as pl
import pyspark.sql as spark
from pyspark.sql import SparkSession

from liken._collections import CollectionsManager
from liken._collections import DeduplicationDict
from liken._dataframe import Frame
from liken._dataframe import wrap
from liken._dedupers import BaseDeduper
from liken._dedupers import exact
from liken._executors import Executor
from liken._executors import LocalExecutor
from liken._executors import SparkExecutor
from liken._pipelines import Pipeline
from liken._types import Columns
from liken._types import DataFrameLike
from liken._types import Keep
from liken._types import UserDataFrame
from liken._validators import validate_columns_arg
from liken._validators import validate_df_arg
from liken._validators import validate_keep_arg
from liken._validators import validate_spark_arg


class Dedupe:
    """Deduplicate a dataframe given a collection of dedupers.

    Apply a deduper as a method call

    Args:
        df: The dataframe to deduplicate.
        spark_session: optional spark session if initializing with PySpark
            backend.

    Returns:
        An Dedupe object with which to "apply" a deduper

    Raises:
        ValueError: Initialized with PySpark DataFrame but no Spark Session.

    Examples:

        import liken as lk

        lk.dedupe(df).apply(exact()).drop_duplicates().collect()
    """

    _executor: Executor

    def __init__(
        self,
        df: UserDataFrame,
        /,
        *,
        spark_session: SparkSession | None = None,
    ):
        self._df: DataFrameLike = validate_df_arg(df)

        self._collection = CollectionsManager()

        if isinstance(df, spark.DataFrame):
            spark_session = validate_spark_arg(spark_session)
            self._executor = SparkExecutor(spark_session=spark_session)
        else:
            self._executor = LocalExecutor()

    @classmethod
    def _from_rows(
        cls,
        rows: list[spark.Row],
    ) -> Dedupe:
        """bypass initialization and initialize explicitely with no validation.

        Use as internal constructor with spark `Rows`.
        """
        self = cls.__new__(cls)
        self._df = rows
        self._collection = CollectionsManager()
        self._executor = LocalExecutor()
        return self

    def apply(self, deduper: BaseDeduper | dict | Pipeline) -> Self:
        """Apply a deduper or dedupers for deduplication.

        Available for inspection when accessed with `.explain()`. Can be
        repetitively called if using the Sequential API. Else apply once using
        the Dict API or Pipeline API.

        Args:
            deduper: The deduper or dedupers to apply

        Returns:
            None

        Raises:
            InvalidDeduperError: For any invalid deduper or collection of
                dedupers

        Example:
            Import and prepate data:

                import liken as lk

            Simple API:

                lk.dedupe(df).apply(lk.exact())

            Dict API:

                lk.dedupe(df).apply({"address": (exact(), tfidf())}

            Pipeline API:

                lk.dedupe(df).apply(
                    lk.pipeline()
                    .step(lk.col("address").exact())
                    .step(lk.col("address").tfidf())
                )

        """
        self._collection.apply(deduper)
        return self

    def drop_duplicates(
        self,
        columns: Columns | None = None,
        *,
        keep: Keep = "first",
    ) -> Self:
        """Drop duplicates by enacting the applied dedupers.

        If no dedupers are explicitely provided, will carry out an exact
        deduplication on any number of columns provided in `columns`.

        Args:
            columns (str | tuple[str, ...] | None): The attribute(s) of the
                dataframe to deduplicate.
            keep: Accepted as "first" or "last". Whether to keep the first intance
                of a duplicate or the last intance, as found in the DataFrame.

        Returns:
            A deduplicated DataFrame.

        Raises:
            ValueError: Incorrect value to `keep` arg.
            ValueError: Incorrect use of `columns` arg given API used to apply
                dedupers.
            ValueError: Incorrect use a single column deduper given multiple
                columns defined, or vice-versa.
        """
        keep: Keep = validate_keep_arg(keep)
        columns: Columns | None = validate_columns_arg(columns, self._collection.is_sequential_applied)
        wdf: Frame = wrap(self._df, None)  # canonical id only ever autoincremental for dropping

        # No .apply(), assumes exact deduplication
        if not self._collection.has_applies:
            self._collection.apply(exact())
        dedupers: DeduplicationDict | Pipeline = self._collection.get()

        self._df: DataFrameLike = self._executor.execute(
            wdf,
            columns=columns,
            dedupers=dedupers,
            keep=keep,
            drop_duplicates=True,
            drop_canonical_id=True,
            id=None,
        ).unwrap()

        self._collection.reset()

        return self

    def canonicalize(
        self,
        columns: Columns | None = None,
        *,
        keep: Keep = "first",
        drop_duplicates: bool = False,
        id: str | None = None,
    ) -> Self:
        """Canonicalize by enacting the applied dedupers.

        If no dedupers are explicitely provided, will carry out an exact
        canonicalization on any number of columns provided in `columns`.

        Args:
            columns (str | tuple[str, ...] | None): The attribute(s) of the
                dataframe to deduplicate.
            keep: Accepted as "first" or "last". Whether to keep the first
                intance of a duplicate or the last intance, as found in the
                DataFrame.
            drop_duplicates: Optionally drop duplicates, whilst preserving a
                canonical_id, contrary to `drop_duplicates`.
            id: string label identifying a column in the dataframe that can be
                used to optionally override the values of a default
                canonical_id.

        Returns:
            A canonicalised DataFrame. By default canonicalization is tracked
                in a new `canonical_id` field.

        Raises:
            ValueError: Incorrect value to `keep` arg.
            ValueError: Incorrect use of `columns` arg given API used to apply
                dedupers.
            ValueError: Incorrect use of a single column deduper given multiple
                columns defined, or vice-versa.
        """
        keep: Keep = validate_keep_arg(keep)
        columns: Columns | None = validate_columns_arg(columns, self._collection.is_sequential_applied)
        wdf: Frame = wrap(self._df, id)

        # No .apply(), assumes exact deduplication
        if not self._collection.has_applies:
            self.apply(exact())
        dedupers: DeduplicationDict | Pipeline = self._collection.get()

        self._df: DataFrameLike = self._executor.execute(
            wdf,
            columns=columns,
            dedupers=dedupers,
            keep=keep,
            drop_duplicates=drop_duplicates,
            drop_canonical_id=False,
            id=id,
        ).unwrap()

        self._collection.reset()

        return self

    def collect(self) -> pd.DataFrame | pl.DataFrame | spark.DataFrame:
        """Collect results and return the dataframe."""
        return self._df

    def explain(self):
        """
        Returns the dedupers as currently stored in the collections manager.

        If no dedupers are stored, returns None. Otherwise, returns a string
        representation of the dedupers collection

        Returns:
            The stored dedupers, formatted

        Examples:

            >>> pipeline = {"address": (lk.exact(), lk.tfidf()), "email": lk.fuzzy()}

            >>> print(lk.dedupe(df).apply(pipeline).explain())

            {
                'address': (
                    exact(),
                    tfidf(threshold=0.95, ngram=3, topn=2),
                    ),
                'email': (
                    fuzzy(threshold=0.95, scorer='simple_ratio'),
                    ),
            }
        """
        return self._collection.pretty_get()


# API:


def dedupe(df: UserDataFrame, /, *, spark_session: SparkSession | None = None) -> Dedupe:
    """Convenience function for `Dedupe` entrypoint."""
    return Dedupe(df, spark_session=spark_session)



"""Define custom dedupers"""

from collections.abc import Iterable
from collections.abc import Iterator
from typing import Callable
from typing import TypeAlias
from typing import final

from typing_extensions import override

from liken._dedupers import ThresholdDedupers
from liken._types import SimilarPairIndices


# TYPES:


PairGenerator: TypeAlias = Callable[[Iterable], Iterable[SimilarPairIndices]]


# CUSTOM:


@final
class Custom(ThresholdDedupers):
    """
    Inherits from Threshold Dedupers for a generalised approach.

    Overrides _gen_similarity_pairs to accept a custom callable, which albeit
    this class being derived from the ThresholdDeduper class, can nevertheless
    be implemented such that it produces predicate results.
    """

    def __init__(
        self,
        pair_fn: PairGenerator,
        /,
        *args,
        **kwargs,
    ):
        super().__init__(
            pair_fn=pair_fn,
            *args,
            **kwargs,
        )
        self._pair_fn = pair_fn
        self._args = args
        self._kwargs = kwargs

    @override
    def validate(self, columns):
        """No validation such that custom can be applied to single or
        compound column
        """
        del columns  # Unused
        pass

    @override
    def _gen_similarity_pairs(self, array) -> Iterator[SimilarPairIndices]:
        """generator or function implementation"""
        array: list = array.to_pylist()
        yield from self._pair_fn(array, *self._args, **self._kwargs)

    def __str__(self):
        return self.__repr__()
